import logging
from collections.abc import Sequence

import nox
from nox.registry import _REGISTRY, Any, Callable, F, Func, Python, functools


def session_decorator(
    func: F | None = None,
    python: Python | None = None,
    py: Python | None = None,
    reuse_venv: bool | None = None,
    name: str | None = None,
    venv_backend: Any | None = None,
    venv_params: Any | None = None,
    tags: Sequence[str] | None = None,
) -> F | Callable[[F], F]:
    """Designate the decorated function as a session."""
    # If `func` is provided, then this is the decorator call with the function
    # being sent as part of the Python syntax (`@nox.session`).
    # If `func` is None, however, then this is a plain function call, and it
    # must return the decorator that ultimately is applied
    # (`@nox.session(...)`).
    #
    # This is what makes the syntax with and without parentheses both work.
    if func is None:
        return functools.partial(
            session_decorator,
            python=python,
            py=py,
            reuse_venv=reuse_venv,
            name=name,
            venv_backend=venv_backend,
            venv_params=venv_params,
            tags=tags,
        )

    if py is not None and python is not None:
        raise ValueError(
            "The py argument to nox.session is an alias for the python "
            "argument, please only specify one."
        )

    if python is None:
        python = py

    final_name = name or func.__name__
    fn = Func(
        func, python, reuse_venv, final_name, venv_backend, venv_params, tags=tags
    )
    _REGISTRY[final_name] = fn
    # this is the line we add to silence the httpx logger
    logging.getLogger("httpx").setLevel(logging.WARNING)
    return fn


nox.session = session_decorator



"""Internal CI tools.

Import the package::

   import laminci

"""

__version__ = "0.11.1"  # denote a pre-release for 0.1.0 with 0.1a1

from . import db, nox
from ._artifacts import upload_docs_artifact
from ._docs import (
    move_built_docs_to_docs_slash_project_slug,
    move_built_docs_to_slash_project_slug,
)
from ._env import get_package_name, get_schema_handle

# backward compat
upload_docs_dir = upload_docs_artifact



from pathlib import Path
from typing import Optional

import toml  # type: ignore
import yaml  # type: ignore


def load_project_yaml(root_directory: Optional[Path] = None) -> dict:
    yaml_file = Path("lamin-project.yaml")
    if root_directory is None:
        root_directory = Path(".")
    yaml_file = root_directory / yaml_file
    with yaml_file.open() as f:
        d = yaml.safe_load(f)
    return d


def get_package_name(root_directory: Optional[Path] = None) -> Optional[str]:
    if Path("lamin-project.yaml").exists():
        config = load_project_yaml(root_directory=root_directory)
        if "package_name" in config:
            return config["package_name"]
        else:
            return None
    elif Path("pyproject.toml").exists():
        with open("pyproject.toml") as f:
            d = toml.load(f)
        return d["project"]["name"].replace("-", "_")
    else:
        return None


def get_schema_handle() -> Optional[str]:
    package_name = get_package_name()
    if package_name is not None:
        if package_name.startswith("lnschema_"):
            return package_name.replace("lnschema_", "")
        else:
            return None
    else:
        raise ValueError(
            "Could not infer python package, add pyproject.toml or update"
            " lamin-project.yaml"
        )



import argparse
import importlib
import os
import subprocess
from subprocess import PIPE, run
from typing import Union

from packaging.version import Version, parse

from ._env import get_package_name

parser = argparse.ArgumentParser("laminci")
subparsers = parser.add_subparsers(dest="command")
migr = subparsers.add_parser(
    "release",
    help="Help with release",
    description=(
        "Assumes you manually prepared the release commit!\n\n"
        "Please edit the version number in your package and prepare the release notes!"
    ),
)
aa = migr.add_argument
aa("--pypi", default=False, action="store_true", help="Publish to PyPI")


def get_last_version_from_tags():
    proc = run(["git", "tag"], universal_newlines=True, stdout=PIPE)
    tags = proc.stdout.splitlines()
    newest = "0.0.0"
    for tag in tags:
        if parse(tag) > parse(newest):
            newest = tag
    return newest


def validate_version(version_str: str):
    version = parse(version_str)
    if version.is_prerelease:
        if not len(version.release) == 2:
            raise SystemExit(
                f"Pre-releases should be of form 0.42a1 or 0.42rc1, yours is {version}"
            )
        else:
            return None
    if len(version.release) != 3:
        raise SystemExit(f"Version should be of form 0.1.2, yours is {version}")


def publish_github_release(
    repo_name: str,
    version: Union[str, Version],
    release_name: str,
    body: str = "",
    draft: bool = False,
    generate_release_notes: bool = True,
):
    version = parse(version)

    try:
        subprocess.run(["gh", "--version"], check=True, stdout=subprocess.PIPE)

        try:
            command = [
                "gh",
                "release",
                "create",
                f"{version}",
                "--title",
                release_name,
                "--notes",
                body,
                "--generate-notes",
            ]
            if version.is_prerelease:
                command.append("--prerelease")

            print(f"\nrun: {command}")
            subprocess.run(command, check=True, stdout=subprocess.PIPE)
        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            raise SystemExit(f"Error creating GitHub release using `gh`: {e}")
    except (subprocess.CalledProcessError, FileNotFoundError):
        try:
            from github import Github, GithubException

            token = (
                os.getenv("GITHUB_TOKEN")
                if os.getenv("GITHUB_TOKEN")
                else input("Github token:")
            )
            g = Github(token)

            try:
                repo = g.get_repo(repo_name)
                repo.create_git_release(
                    tag=str(version),
                    name=release_name,
                    message=body,
                    draft=draft,
                    prerelease=version.is_prerelease,
                    generate_release_notes=generate_release_notes,
                )
            except GithubException as e:
                raise SystemExit(f"Error creating GitHub release using `PyGithub`: {e}")
        except ImportError:
            raise SystemExit(
                "Neither the Github CLI ('gh') nor PyGithub were accessible.\n"
                "Please install one of the two."
            )


def main():
    args = parser.parse_args()

    if args.command == "release":
        package_name = get_package_name()
        # cannot do the below as this wouldn't register immediate changes
        # from importlib.metadata import version as get_version
        # version = get_version(package_name)
        module = importlib.import_module(package_name, package=".")
        version = module.__version__
        previous_version = get_last_version_from_tags()
        validate_version(version)
        if parse(version) <= parse(previous_version):
            raise SystemExit(
                f"Your version ({version}) should increment the previous version"
                f" ({previous_version})"
            )

        pypi = " & publish to PyPI" if args.pypi else ""
        response = input(f"Bump {previous_version} to {version}{pypi}? (y/n)")
        if response != "y":
            return None

        commands = [
            "git add -u",
            f"git commit -m 'Release {version}'",
            "git push",
            f"git tag {version}",
            f"git push origin {version}",
        ]
        for command in commands:
            print(f"\nrun: {command}")
            run(command, shell=True)

        publish_github_release(
            repo_name=f"laminlabs/{package_name}",
            version=version,
            release_name=f"Release {version}",
        )

        if args.pypi:
            command = "flit publish"
            print(f"\nrun: {command}")
            run(command, shell=True)
